'use strict';

const EventEmitter = require('events');
const Promise      = require('bluebird');
const debug        = require('debug')('eventbus:main');
const uuid         = require('uuid');
const CFError      = require('cf-errors');
const Bus          = require('./Bus');
const Store        = require('./Store');

//TODO
// 1. queue published messages in case of failure in the connection
// 2. re subscribe in case of reconnection
// 3. allow to control how and when to ack a subscribed event
// 4. add ability to get all possible event types in the system

class EventBus extends EventEmitter {

    constructor() {
        super();
        this.busReady     = false;
        this.storeReady   = false;
        this.readyPromise = new Promise((resolve, reject) => {
            this.readyDeferred = {
                resolve: () => {
                    resolve();
                },
                reject: () => {
                    reject();
                }
            };
        });
        this.on('error', () => {});
    }

    async quit() {
        try {
            await Promise.all([this.bus.quit(), this.store.quit()]);
            this.emit('quit');
        } catch (err) {
            throw new CFError({
                cause: err,
                message: 'Failed during quit execution'
            });
        }
    }

    init(options = {}) {
        this.microServiceName = options.microServiceName || `eventBus_client_${uuid()}`;

        this.bus   = new Bus(this.microServiceName, options.bus);
        this.store = new Store(this.microServiceName, options.store);

        // wait for both the bus and the store to send their ready event
        this.bus.on('ready', () => {
            debug('bus is ready. registered as: ${this.microServiceName}`');
            this.busReady = true;
            this._emitReady();
        });

        this.store.on('ready', () => {
            debug(`store is ready. registered as: ${this.microServiceName}`);
            this.storeReady = true;
            this._emitReady();
        });

        this.bus.on('error', (err) => {
            debug(`bus error: ${err.toString()}`);
            this.emit('error', new CFError({
                cause: err,
                message: 'eventbus bus error'
            }));
        });

        this.store.on('error', (err) => {
            debug(`store error: ${err.toString()}`);
            this.emit('error', new CFError({
                cause: err,
                message: 'eventbus store error'
            }));
        });

        this.bus.init();
        this.store.init();
    }

    _emitReady() {
        if (this.busReady && this.storeReady) {
            this.readyDeferred.resolve();
            debug('ready');
            this.emit('ready');
        }
    }

    async publish(eventName, msg) {
        await this.readyPromise;
        await this.store.save(eventName, msg);
        await this.bus.publish(eventName, msg);
    }

    async subscribe(eventName, handler) {
        await this.readyPromise;
        await this.bus.subscribe(eventName, handler);
    }

    /**
     * gets all the events that matches an aggregate id
     * @param aggregateId
     * @returns pg extended Cursor
     */
    async getAllAggregateEvents(aggregateId) {
        await this.readyPromise;
        return await this.store.getAllAggregateEvents(aggregateId);
    }

    /**
     * gets all the events that matches the provided patterns
     * @param name
     * @returns pg extended Cursor
     */
    async getAllEventsByName() {
        await this.readyPromise;
        return await this.store.getAllEventsByName(...arguments);
    }

}

module.exports = new EventBus();
